""" Module implementing the "Scheduled Jobs" functionality of Confluence Server
"""

import dataclasses
import typing

import requests
from bs4 import BeautifulSoup
from furl import furl

LIST_JOBS_ACTION_URL = "/admin/scheduledjobs/viewscheduledjobs.action"
RUN_JOB_ACTION_URL = "/admin/scheduledjobs/runJob.action"
DISABLE_JOB_ACTION_URL = "/admin/scheduledjobs/disableJob.action"
ENABLE_JOB_ACTION_URL = "/admin/scheduledjobs/enableJob.action"


@dataclasses.dataclass()
class Job:
    name: str
    group: str
    id: str
    status: str
    last_execution: typing.Optional[str]
    next_execution: typing.Optional[str]
    avg_duration: typing.Optional[int]
    has_history: bool
    is_runnable: bool
    is_editable: bool
    action_enable_disable: typing.Optional[str]
    is_cron: bool
    cron_expression: typing.Optional[str]
    repeat_interval: typing.Optional[str]


def get_token_and_cookies(base_url: furl) -> typing.Tuple[str, requests.cookies.RequestsCookieJar]:
    request_url = base_url.copy()
    request_url.add(path=LIST_JOBS_ACTION_URL)
    response = requests.get(request_url)

    soup = BeautifulSoup(response.content, "html.parser")

    token = soup.select_one("meta#atlassian-token")["content"]

    return (token, response.cookies)


def list_jobs(base_url: furl) -> typing.Tuple[typing.List[Job], str, requests.cookies.RequestsCookieJar]:
    """
    Returns:
        - list of jobs (typing.List[Job])
        - token(str)
        - cookies of request(requests.cookies.RequestsCookieJar)
    """
    request_url = base_url.copy()
    request_url.add(path=LIST_JOBS_ACTION_URL)
    response = requests.get(request_url)

    soup = BeautifulSoup(response.content, "html.parser")

    token = soup.select_one("meta#atlassian-token")["content"]
    table = soup.select_one("table#schedule-admin")
    headers = [x.text for x in table.select("thead th")]
    table_rows = table.select("tbody tr")

    entries: typing.List[Job] = []
    for row in table_rows:
        job_name = row["data-job-name"]
        job_group = row["data-job-group"]
        job_id = row["data-job-id"]
        is_cron = row["data-is-cron"]
        cron_expression = row["data-cron-expression"]
        repeat_interval = row["data-repeat-interval"]
        has_history = True if row.select_one(".show-history") is not None else False
        is_runnable = True if row.select_one(".run-job") is not None else False
        is_editable = True if row.select_one(".edit-schedule") is not None else False
        has_disable = True if row.select_one(".disable-job") is not None else False
        has_enable = True if row.select_one(".enable-job") is not None else False

        status = row.select("td")[headers.index("Status")].get_text()
        last_execution = row.select("td")[headers.index("Last Execution")].get_text()
        next_execution = row.select("td")[headers.index("Next Execution")].get_text()
        avg_duration = row.select("td")[headers.index("Avg. Duration")].get_text()

        action_enable_disable = None
        if has_disable:
            action_enable_disable = "disable"
        elif has_enable:
            action_enable_disable = "enable"
        entries.append(
            Job(
                name=job_name,
                group=job_group,
                id=job_id,
                status=status,
                last_execution=last_execution,
                next_execution=next_execution,
                avg_duration=avg_duration,
                has_history=has_history,
                is_runnable=is_runnable,
                is_editable=is_editable,
                action_enable_disable=action_enable_disable,
                is_cron=is_cron,
                cron_expression=cron_expression,
                repeat_interval=repeat_interval,
            )
        )

    return (entries, token, response.cookies)


def run_job(
    base_url: furl,
    job: Job,
    token: typing.Optional[str] = None,
    cookies: typing.Optional[requests.cookies.RequestsCookieJar] = None,
) -> bool:
    if token is None or cookies is None:
        token, cookies = get_token_and_cookies(base_url)
    request_url = base_url.copy()
    request_url.add(path=RUN_JOB_ACTION_URL)
    request_url.add(args={"group": job.group, "id": job.id, "atl_token": token})
    response = requests.get(request_url, cookies=cookies)
    return response.status_code == 200


def disable_job(
    base_url: furl,
    job: Job,
    token: typing.Optional[str] = None,
    cookies: typing.Optional[requests.cookies.RequestsCookieJar] = None,
) -> bool:
    if token is None or cookies is None:
        token, cookies = get_token_and_cookies(base_url)
    request_url = base_url.copy()
    request_url.add(path=DISABLE_JOB_ACTION_URL)
    request_url.add(args={"group": job.group, "id": job.id, "atl_token": token})
    response = requests.get(request_url, cookies=cookies)
    return response.status_code == 200


def enable_job(
    base_url: furl,
    job: Job,
    token: typing.Optional[str] = None,
    cookies: typing.Optional[requests.cookies.RequestsCookieJar] = None,
) -> bool:
    if token is None or cookies is None:
        token, cookies = get_token_and_cookies(base_url)
    request_url = base_url.copy()
    request_url.add(path=ENABLE_JOB_ACTION_URL)
    request_url.add(args={"group": job.group, "id": job.id, "atl_token": token})
    response = requests.get(request_url, cookies=cookies)
    return response.status_code == 200



""" This module provides some basic path tools for the pluploader cli tool
"""

import os
import pathlib
import typing
import xml.etree.ElementTree as ET


class PluginKeyNotFoundError(RuntimeError):
    pass


def get_jar_path_from_pom() -> os.PathLike:
    """ Get jar to upload based on maven pom

    This function reads the pom and analyses which artifact was build by the last
    `mvn package` command. If the file exists, the file will be returned
    """
    rootdir = find_maven_project_root()
    namespace = {"ns": "http://maven.apache.org/POM/4.0.0"}

    root = ET.parse(f"{rootdir}/pom.xml").getroot()
    artifact_id = root.find("ns:artifactId", namespace).text
    version = root.find("ns:version", namespace).text

    return rootdir / "target" / f"{artifact_id}-{version}.jar"


def get_plugin_key_from_pom() -> str:
    """ Get Plugin key from Pom.xml

    This function reads the pom and analyses which plugin will be built.
    """
    try:
        rootdir = find_maven_project_root()
        namespace = {"ns": "http://maven.apache.org/POM/4.0.0"}
        root = ET.parse(f"{rootdir}/pom.xml").getroot()
        properties = root.find("ns:properties", namespace)
        plugin_id = properties.find("ns:atlassian.plugin.key", namespace).text
        return plugin_id
    except FileNotFoundError as exc:
        raise exc
    except Exception as exc:
        raise PluginKeyNotFoundError(exc)


def find_maven_project_root(working_path: pathlib.Path = pathlib.Path(".")) -> pathlib.Path:
    """Tries to find a maven project root directory.

    Tries to find a maven project root directory if the current path is a
    parent directory. Works by finding the pom.xml file.
    Args:
        working_path: a string representation of the directory you want to find
            the project
    Returns:
        the absolute project path
    """
    for walk_tuple in _walk_up(working_path):
        if "pom.xml" in walk_tuple[2]:
            return walk_tuple[0]
    raise FileNotFoundError()


def _walk_up(
    start_path: pathlib.Path = pathlib.Path(),
) -> typing.Tuple[pathlib.Path, typing.Tuple[os.PathLike], typing.Tuple[os.PathLike]]:
    """ Generator for walking up a file path. os.walk like behavior

    Args: start_path: a os.PathLike path to start from

    Yields:
        3-Tuple (dirpath, dirnames, filenames)
    """
    current_path = start_path.resolve()
    while True:
        walk_tuple = (
            current_path,
            [x.name for x in current_path.resolve().iterdir() if x.is_dir()],
            [x.name for x in current_path.resolve().iterdir() if not x.is_dir()],
        )
        yield walk_tuple
        if current_path == pathlib.Path(current_path.root):
            return
        current_path = current_path.parent



""" This module provides a basic interface for the upm rest api
"""

import dataclasses
import json
import typing

import requests
from colorama import Fore
from furl import furl
from packaging import version

UPM_API_ENDPOINT: str = "/rest/plugins/1.0/"


@dataclasses.dataclass()
class ModuleDto:
    completeKey: str
    key: typing.Optional[str]
    name: str
    enabled: bool
    optional: bool
    recognisableType: bool
    broken: bool

    @classmethod
    def decode(cls, obj: dict) -> "ModuleDto":
        if "key" in obj:
            return cls(
                key=obj.get("key"),
                completeKey=obj.get("completeKey", ""),
                name=obj.get("name", ""),
                enabled=obj.get("enabled", False),
                optional=obj.get("optional", False),
                recognisableType=obj.get("recognisableType", False),
                broken=obj.get("broken", True),
            )
        raise ValueError("decode expected passed object to have a key.")


@dataclasses.dataclass
class PluginDto:
    """ This class represents a plugin given by the UPM/Plugin API
    """

    key: str
    name: str
    version: version.Version
    enabled: bool
    userInstalled: bool
    description: str
    modules: typing.Optional[typing.List[ModuleDto]]

    def print_table(self, print_modules: bool):
        """Prints table view of plugin information
        """
        for key, value in self.__dict__.items():
            if key == "modules":
                if print_modules:
                    print(f"{(key + ':'):20}")
                    for module in value:
                        if module.enabled:
                            status = f"{Fore.GREEN}✓{Fore.RESET}"
                        else:
                            status = f"{Fore.YELLOW}!{Fore.RESET}"
                        print(f"  {status} {module.name[:20]:20} {module.key}")
            else:
                print(f"{(key + ':'):15} {value}")

    @classmethod
    def decode(cls, obj: dict) -> "PluginDto":
        if "key" in obj:
            return cls(
                key=obj.get("key"),
                name=obj.get("name", ""),
                version=obj.get("version", "0.0.1"),
                userInstalled=obj.get("userInstalled", False),
                enabled=obj.get("enabled", False),
                description=obj.get("description", ""),
                modules=[ModuleDto.decode(x) for x in obj.get("modules", [])],
            )
        raise ValueError("decode expected passed object to have a key.")


def get_token(base_url: furl) -> str:
    """ Get token from api endpoint
    """
    token_url: furl = base_url.copy()
    token_url.add(path=UPM_API_ENDPOINT)
    token_url.set(args={"os_authType": "basic"})
    token_response = requests.head(token_url.url)
    token = token_response.headers["upm-token"]
    return token


def upload_plugin(base_url: furl, files: typing.Dict[str, typing.BinaryIO], token: str) -> typing.Tuple[int, typing.Any]:
    """ Upload plugin
    """
    upload_url = base_url.copy()
    upload_url.set(args={"token": token})
    upload_url.add(path=UPM_API_ENDPOINT)
    upload_response = requests.post(upload_url.url, files=files)
    text = upload_response.text.replace("<textarea>", "").replace("</textarea>", "")
    upload_response_data = json.loads(text)
    progress = int(upload_response_data.get("status", {}).get("amountDownloaded", 0))
    return progress, upload_response_data


def get_current_progress(base_url: furl, previous_request) -> typing.Tuple[int, typing.Dict]:
    progress_url = base_url.copy()
    progress_url.set(path=previous_request["links"]["self"])
    progress_rd = requests.get(progress_url.url).json()
    if "type" in progress_rd:
        progress = int(progress_rd.get("status", {}).get("amountDownloaded", 0))
        return progress, progress_rd
    return 100, progress_rd


def get_all_plugins(base_url: furl, user_installed: bool = True) -> typing.List[PluginDto]:
    """ Gets a list of all installed plugins from the api and returns it
    If user_installed is set true (default), only user installed plugins are listed
    """
    request_url = base_url.copy()
    request_url.add(path=UPM_API_ENDPOINT)
    response = requests.get(request_url.url)
    return_obj = [PluginDto.decode(x) for x in response.json().get("plugins", [])]
    if user_installed:
        return_obj = filter(lambda x: x.userInstalled, return_obj)
    return return_obj


def get_plugin(base_url: furl, plugin_key: str) -> PluginDto:
    """ Gets Plugin info by using the UPM_API_ENDPOINT/plugin-key/ endpoint and
    returns it as a PluginDto
    """
    request_url = base_url.copy()
    request_url.add(path=UPM_API_ENDPOINT)
    request_url.join(plugin_key + "-key")
    response = requests.get(request_url.url)
    return_obj = PluginDto.decode(response.json())
    return return_obj


def enable_disable_plugin(base_url: furl, plugin_key: str, enabled: bool) -> PluginDto:
    """ Enables/Disables Plugin"""
    mod = {"enabled": enabled}
    return _modify_plugin(base_url, plugin_key, mod)


def _modify_plugin(base_url: furl, plugin_key: str, modifications: dict) -> PluginDto:
    """ Puts Changes to plugin by using the UPM_API_ENDPOINT/plugin-key/ endpoint and
    returns new infos as a PluginDto
    """
    request_url = base_url.copy()
    request_url.add(path=UPM_API_ENDPOINT)
    request_url.join(plugin_key + "-key")
    headers = {"Content-Type": "application/vnd.atl.plugins.plugin+json"}
    response = requests.put(request_url.url, json=modifications, headers=headers)
    return_obj = PluginDto.decode(response.json())
    return return_obj


def uninstall_plugin(base_url: furl, plugin_key: str) -> bool:
    """ Uninstalls a plugin by using the UPM_API_ENDPOINT/plugin-key/ endpoint
    """
    request_url = base_url.copy()
    request_url.add(path=UPM_API_ENDPOINT)
    request_url.join(plugin_key + "-key")
    response = requests.delete(request_url.url)
    return response.status_code == 204


def module_status(previous_request: dict,) -> typing.Tuple[int, int, typing.List[ModuleDto]]:
    """ Returns the module status of an plugin based on a request/dict containing an PluginDto
    returns an tuple containing
        1. number of all plugins
        2. number of enabled plugins
        3. Array of disabled plugins
    """
    plugin_dto = PluginDto.decode(previous_request)
    disabled_modules = [module for module in plugin_dto.modules if not module.enabled]
    all_modules = len(plugin_dto.modules)
    enabled_modules = all_modules - len(disabled_modules)
    return all_modules, enabled_modules, disabled_modules


def get_safemode(base_url: furl) -> bool:
    request_url = base_url.copy()
    request_url.add(path=UPM_API_ENDPOINT)
    request_url.add(path="safe-mode")
    response = requests.get(request_url.url)
    return response.json()["enabled"]


def enable_disable_safemode(base_url: furl, enable: bool, keepState: bool = False) -> bool:
    headers = {"Content-Type": "application/vnd.atl.plugins.safe.mode.flag+json"}
    data = {
        "enabled": enable,
        "links": {},
    }

    request_url = base_url.copy()
    request_url.add(path=UPM_API_ENDPOINT)
    request_url.add(path="safe-mode")
    request_url.add(query_params={"keepState": "true" if keepState else "false"})
    response = requests.put(request_url.url, headers=headers, json=data)
    response_json = response.json()
    return "subCode" not in response_json and response_json["enabled"] == enable



""" this module helps interacting with .jar-files built working for atlassian
server applications.
"""

import pathlib
from zipfile import ZipFile
from bs4 import BeautifulSoup
from pluploader.pathutil import PluginKeyNotFoundError


def _get_atlassian_plugin_xml_from_jar_path(path: pathlib.Path) -> str:
    """ Opens the jar on the provided path and tries to find the
    atlassian-plugin.xml in this file
    Args:
        path (pathlib.Param): the path to the jar file
    Returns:
        str: the content of atlassian_plugin.xml

    Raises:
        FileNotFoundError: If the file of path is not found
        zipfile.BadZipFile: If the provided file of path is not a zip file
        KeyError: If no atlassian_plugin.xml is existing inside the zip/jar
    """
    with ZipFile(path) as jar:
        with jar.open("atlassian-plugin.xml") as atlassian_plugin_xml:
            return atlassian_plugin_xml.read()


def _find_plugin_key(atlassian_plugin_xml: str) -> str:
    """ Finds the plugin key in an atlassian_plugin_xml
    Args:
        atlassian_plugin_xml: the content of an atlassian_plugin.xml
    Returns:
        str: the plugin key of the provided atlassian_plugin.xml
    Raises:
        pluploader.pathutil.PluginKeyNotFoundError: If the PluginKey could not be found
    """
    soup = BeautifulSoup(atlassian_plugin_xml, "xml")
    key = soup.find("atlassian-plugin").get("key")
    if key is None:
        raise PluginKeyNotFoundError()
    return key


def get_plugin_key_from_jar_path(path: pathlib.Path) -> str:
    """ Tries to find the plugin key of an atlassian server app plugin by providing an path to the jar.
    Args:
        path (pathlib.Param): the path to the jar file
    Returns:
        str: The Plugin Key
    Raises:
        FileNotFoundError: If the file of path is not found
        zipfile.BadZipFile: If the provided file of path is not a zip file
        KeyError: If no atlassian_plugin.xml is existing inside the zip/jar
        pluploader.pathutil.PluginKeyNotFoundError: If the PluginKey could not be found
    """
    atlas_xml = _get_atlassian_plugin_xml_from_jar_path(path)
    return _find_plugin_key(atlas_xml)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pluploader']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.3.1,<6.0.0',
 'beautifulsoup4>=4.9.3,<5.0.0',
 'click_default_group>=1.2.2,<2.0.0',
 'colorama>=0.4.3,<0.5.0',
 'coloredlogs>=14.0,<15.0',
 'configargparse>=1.2.3,<2.0.0',
 'furl>=2.1.0,<3.0.0',
 'importlib-metadata>=1.7.0,<2.0.0',
 'lxml>=4.5.2,<5.0.0',
 'packaging>=20.4,<21.0',
 'pydantic>=1.6.1,<2.0.0',
 'requests>=2.24.0,<3.0.0',
 'tqdm>=4.48.2,<5.0.0',
 'typer>=0.3.2,<0.4.0']

extras_require = \
{':python_version < "3.7"': ['dataclasses']}

entry_points = \
{'console_scripts': ['pluploader = pluploader.main:main']}

setup_kwargs = {
    'name': 'pluploader',
    'version': '0.5.0b1',
    'description': 'CLI Confluence/Jira Plugin uploader',
    'long_description': None,
    'author': 'Fabian Siegel',
    'author_email': 'fabian@livelyapps.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



""" pluploader executable
"""
import logging
import pathlib
import shutil
import sys
import time
import typing

import coloredlogs
import furl
import requests
import typer
import yaml
from click_default_group import DefaultGroup
from colorama import Fore
from tqdm import tqdm
import zipfile

from . import __version__, jobs, pathutil
from . import atlas_jar_util as jar
from . import upmapi as upm

app = typer.Typer()
app_safemode = typer.Typer()
app_job = typer.Typer()

app.add_typer(app_safemode, name="safe-mode")
app.add_typer(app_job, name="job")

LOGO = f"""
{Fore.YELLOW} ))))          {Fore.RED}           ((((
{Fore.YELLOW} )))))))       {Fore.RED}        (((((((
{Fore.YELLOW}  ))))))))     {Fore.RED}      ((((((((
{Fore.YELLOW}  ))))))))))   {Fore.RED}    ((((((((((
{Fore.YELLOW}   ))))))))))) {Fore.RED}  (((((((((((
{Fore.YELLOW}    ))))))))))){Fore.GREEN}|{Fore.RED}(((((((((((
{Fore.YELLOW}     ))))))))){Fore.GREEN}|||{Fore.RED}(((((((((
{Fore.YELLOW}      ))))))){Fore.GREEN}|||||{Fore.RED}(((((((
{Fore.YELLOW}        )))){Fore.GREEN}|||||||{Fore.RED}((((
{Fore.YELLOW}          )){Fore.GREEN}|||||||{Fore.RED}((
{Fore.YELLOW}            {Fore.RED}/{Fore.GREEN}|||||{Fore.YELLOW}\\
{Fore.RESET}"""

coloredlogs.install(level="DEBUG")
coloredlogs.install(fmt="%(asctime)s %(levelname)s %(message)s")


def main():
    """ Reads config and passes it to app
    """
    config_locations = []
    home_cfg = pathlib.Path().home() / pathlib.Path(".pluprc")
    if home_cfg.exists():
        config_locations.append(home_cfg)
    try:
        project_root = pathutil.find_maven_project_root()
        project_cfg = project_root / ".pluprc"
        if project_cfg.exists():
            config_locations.append(project_cfg)
    except FileNotFoundError:
        pass
    settings = {}

    for config_location in config_locations:
        with open(config_location) as stream:
            try:
                settings.update(yaml.safe_load(stream))
            except yaml.YAMLError:
                logging.warning("Looks like your configuration file is not yaml, the file will be ignored")

    cmd: DefaultGroup = typer.main.get_command(app)
    cmd.default_if_no_args = True
    cmd.default_cmd_name = "install"
    cmd.context_settings = {"default_map": settings}
    cmd()


def version_callback(value: bool):
    if value:
        print(f"You're using {__version__}")
        raise typer.Exit()


def furl_callback(value: str) -> furl.furl:
    try:
        _url = furl.furl(value)
        if not (furl.is_valid_host(_url.host) and furl.is_valid_port(_url.port) and furl.is_valid_scheme(_url.scheme)):
            raise ValueError
    except Exception:
        raise typer.BadParameter('Make sure to provide an valid url (e.G. "https://your.confluence.net:8090")')
    return _url


@app.callback(cls=DefaultGroup)
def root(
    ctx: typer.Context,
    version: typing.Optional[bool] = typer.Option(False, "--version", callback=version_callback, is_eager=True,),
    base_url: str = typer.Option(
        "http://localhost:8090",
        help="Set the base-url of your instance. This flag will overwrite scheme, host, path and port, if those are set in"
        "this string.",
        callback=furl_callback,
    ),
    user: str = typer.Option("admin", help="Set the username of the user you want to use",),
    password: str = typer.Option("admin", help="Set the password of the user you want to use",),
    port: typing.Optional[int] = typer.Option(None),
    ask_for_password: typing.Optional[bool] = typer.Option(False, help="Asks user for password interactively"),
    logo: bool = typer.Option(True, help="Print lively apps logo"),
):
    """ A simple command line plugin uploader/installer/manager for atlassian product server
    instances (Confluence/Jira) written in python(3).
    """
    if logo:
        print(LOGO)
    if ask_for_password:
        password = typer.prompt("Password: ", hide_input=True)
    burl: furl.furl = _base_url_from_args(base_url, user, password, port)
    ctx.obj = {"base_url": burl}


def _base_url_from_args(base_url: str, user: str, password: str, port: typing.Optional[int]) -> furl.furl:
    """creates furl instance from defaults, config(via defaults) and args
    """
    base_url.username = user
    base_url.password = password
    if port is not None:
        base_url.port = port
    return base_url


@app.command("list")
def list_all(
    ctx: typer.Context, print_all: bool = typer.Option(False, help="prints all plugins instead of only user installed plugins")
):
    """ Prints out basic plugin informations of all plugins"""
    try:
        all_plugins = upm.get_all_plugins(ctx.obj.get("base_url"), not print_all)
    except requests.exceptions.ConnectionError:
        logging.error("Could not connect to host - check your base-url")
        sys.exit(1)
    except Exception as exc:
        logging.error("An error occured - check your credentials")
        logging.error("%s", exc)
        sys.exit(1)
    print(f"  {'Name':25} {'Version':13} {'Plugin Key':50}")
    for plugin in all_plugins:
        if plugin.enabled:
            status = f"{Fore.GREEN}✓{Fore.RESET}"
        else:
            status = f"{Fore.YELLOW}!{Fore.RESET}"

        plugin_infos = f"{status} {plugin.name[:25]:25}" + f" {str(plugin.version)[:13]:13} ({plugin.key})"
        print(plugin_infos)


@app.command("info")
def plugin_info(
    ctx: typer.Context,
    plugin: str = typer.Argument(None, help="the plugin key"),
    show_modules: bool = typer.Option(False, help="show modules of plugin as well"),
):
    """ prints information of the plugin specified by the plugin key
    """
    if plugin is None:
        try:
            plugin = pathutil.get_plugin_key_from_pom()
        except FileNotFoundError:
            logging.error("Could not find the plugin you want to get the info of. Are you in a maven directory?")
            sys.exit(1)
        except pathutil.PluginKeyNotFoundError:
            logging.error("Could not find the plugin you want to get the info of. Is the plugin key set in the pom.xml?")
            sys.exit(1)
    try:
        info = upm.get_plugin(ctx.obj.get("base_url"), plugin)
    except requests.exceptions.ConnectionError:
        logging.error("Could not connect to host - check your base-url")
        sys.exit(1)
    except Exception as exc:
        logging.error("An error occured - check your credentials")
        logging.error("%s", exc)
        sys.exit(1)
    info.print_table(show_modules)


@app.command("enable")
def enable_plugin(
    ctx: typer.Context, plugin: str = typer.Argument(None, help="the plugin key"),
):
    """ Enables the specified plugin """
    if plugin is None:
        try:
            plugin = pathutil.get_plugin_key_from_pom()
        except FileNotFoundError:
            logging.error("Could not find the plugin you want to get the info of. Are you in a maven directory?")
            sys.exit(1)
        except pathutil.PluginKeyNotFoundError:
            logging.error("Could not find the plugin you want to get the info of. Is the plugin key set in the pom.xml?")
            sys.exit(1)
    try:
        response = upm.enable_disable_plugin(ctx.obj.get("base_url"), plugin, True)
    except requests.exceptions.ConnectionError:
        logging.error("Could not connect to host - check your base-url")
        sys.exit(1)
    except Exception as exc:
        logging.error("An error occured - check your credentials")
        logging.error("%s", exc)
        sys.exit(1)
    response.print_table(False)


@app.command("disable")
def disable_plugin(
    ctx: typer.Context, plugin: str = typer.Argument(None, help="the plugin key"),
):
    """ Disables the specified plugin """
    if plugin is None:
        try:
            plugin = pathutil.get_plugin_key_from_pom()
        except FileNotFoundError:
            logging.error("Could not find the plugin you want to get the info of. Are you in a maven directory?")
            sys.exit(1)
        except pathutil.PluginKeyNotFoundError:
            logging.error("Could not find the plugin you want to get the info of. Is the plugin key set in the pom.xml?")
            sys.exit(1)
    try:
        response = upm.enable_disable_plugin(ctx.obj.get("base_url"), plugin, False)
    except requests.exceptions.ConnectionError:
        logging.error("Could not connect to host - check your base-url")
        sys.exit(1)
    except Exception as exc:
        logging.error("An error occured - check your credentials")
        logging.error("%s", exc)
        sys.exit(1)
    response.print_table(False)


@app.command("uninstall")
def uninstall_plugin(
    ctx: typer.Context, plugin: str = typer.Argument(None, help="the plugin key"),
):
    """ Uninstalls a plugin
    """
    if plugin is None:
        try:
            plugin = pathutil.get_plugin_key_from_pom()
        except FileNotFoundError:
            logging.error("Could not find the plugin you want to get the info of. Are you in a maven directory?")
            sys.exit(1)
        except pathutil.PluginKeyNotFoundError:
            logging.error("Could not find the plugin you want to get the info of. Is the plugin key set in the pom.xml?")
            sys.exit(1)
    try:
        status = upm.uninstall_plugin(ctx.obj.get("base_url"), plugin)
    except requests.exceptions.ConnectionError:
        logging.error("Could not connect to host - check your base-url")
        sys.exit(1)
    except Exception as exc:
        logging.error("An error occured - check your credentials")
        logging.error("%s", exc)
        sys.exit(1)
    if status:
        logging.info("Plugin successfully uninstalled")
    else:
        logging.error("An error occurred. The plugin could not be uninstalled.")


@app.command("install")
def install(
    ctx: typer.Context,
    file: typing.Optional[pathlib.Path] = typer.Option(None, "--file", "-f", help=""),
    interactive: typing.Optional[bool] = typer.Option(
        False,
        "--interactive",
        "-i",
        help="Pluploader tries find an plugin in the current directory. If you want to specify the location of the plugin "
        "you want to upload, use -f /path/to/jar",
    ),
    reinstall: typing.Optional[bool] = typer.Option(
        False, "--reinstall", help="Plugin will be uninstalled before it will be installed"
    ),
):
    """installs the plugin of the current maven project or a specified one; you can also omit install
    """
    base_url: furl.furl = ctx.obj.get("base_url")
    if file is None:
        try:
            plugin_path = pathutil.get_jar_path_from_pom()
        except FileNotFoundError:
            logging.error("Could not find the plugin you want to install. Are you in a maven directory?")
            sys.exit(1)
    else:
        plugin_path = file

    displayed_base_url = base_url.copy().remove(username=True, password=True)

    if interactive:
        confirm = input("Do you really want to upload and install the plugin? (y/N) ")
        if confirm.lower() != "y":
            sys.exit()

    if reinstall:
        try:
            plugin_key = jar.get_plugin_key_from_jar_path(plugin_path)
            try:
                status = upm.uninstall_plugin(base_url, plugin_key)
            except requests.exceptions.ConnectionError:
                logging.error("Could not connect to host - check your base-url")
                sys.exit(1)
            except Exception as exc:
                logging.error("An error occured - check your credentials")
                logging.error("%s", exc)
                sys.exit(1)
            if status:
                logging.info("Plugin successfully uninstalled")
            else:
                logging.error("An error occurred. The plugin could not be uninstalled.")
        except (FileNotFoundError, zipfile.BadZipFile, KeyError, pathutil.PluginKeyNotFoundError):
            logging.error("Could not get the plugin key of the supplied jar - are you sure you want to upload a plugin, mate?")

    logging.info(f"{pathlib.Path(plugin_path).name} will be uploaded to {displayed_base_url}")

    try:
        token = upm.get_token(base_url)
    except requests.exceptions.RequestException:
        logging.error("Could not connect to host - check your base-url")
        sys.exit(1)
    except KeyError:
        logging.error("UPM Token couldn't be retrieved; are your credentials correct?")
        sys.exit(1)

    try:
        with open(plugin_path, "rb") as plugin_file:
            files = {"plugin": plugin_file}
            with TqdmUpTo(total=100) as pbar:
                pbar.update_to(0)
                progress, previous_request = upm.upload_plugin(base_url, files, token)
                while progress != 100:
                    progress, previous_request = upm.get_current_progress(base_url, previous_request)
                    pbar.update_to(progress)
                    time.sleep(0.1)
    except requests.exceptions.RequestException:
        logging.error("An error occured while uploading plugin")
        sys.exit(1)
    except FileNotFoundError:
        logging.error("Could not find the plugin you want to install.")
        sys.exit(1)
    finally:
        for file in files.values():
            file.close()

    if previous_request["enabled"]:
        status = f"{Fore.GREEN}enabled{Fore.RESET}!"
    else:
        status = (
            f"{Fore.RED}disabled{Fore.RESET}! \n"
            "You should check the logs of your Atlassian host to find out why your plugin was disabled."
        )

    all_nr, enabled, disabled = upm.module_status(previous_request)
    logging.info("plugin uploaded and " + status + f" ({enabled} of {all_nr} modules enabled)")
    if len(disabled) != 0 and len(disabled) != all_nr:
        for module in disabled:
            logging.info(f"   - {module.key} is disabled")
    elif len(disabled) == all_nr:
        logging.error(
            "Your plugin was installed successfully but all modules are disabled. This is often caused by problems such as"
            " importing services that are not properly defined in your atlassian-plugin.xml."
        )
        logging.error("Check the logs of your Atlassian host to find out more.")


@app_safemode.callback()
def safemode(ctx: typer.Context):
    """ Controls the upm safemode
    """


@app_safemode.command("status")
def safemode_status(ctx: typer.Context):
    """ prints out the safemode status """
    try:
        safemode_st = (
            f"{Fore.YELLOW}enabled{Fore.RESET}"
            if upm.get_safemode(ctx.obj.get("base_url"))
            else f"{Fore.GREEN}disabled{Fore.RESET}"
        )
        logging.info("Safe-mode is currently %s", safemode_st)
    except requests.exceptions.ConnectionError:
        logging.error("Could not connect to host - check your base-url")
        sys.exit(1)
    except Exception as e:
        logging.error("An error occured - check your credentials")
        logging.error(f"{e}")
        sys.exit(1)


@app_safemode.command("enable")
def safemode_enable(ctx: typer.Context):
    try:
        success = upm.enable_disable_safemode(ctx.obj.get("base_url"), True)
        if success:
            logging.info(f"Safe-mode is now {Fore.GREEN}enabled{Fore.RESET}")
        else:
            logging.error("Could not enable safe-mode - is safe-mode already enabled?")
    except requests.exceptions.ConnectionError:
        logging.error("Could not connect to host - check your base-url")
        sys.exit(1)
    except Exception as exc:
        logging.error("An error occured - check your credentials")
        logging.error("%s", exc)
        sys.exit(1)


@app_safemode.command("disable")
def safemode_disable(ctx: typer.Context, keep_state: bool = typer.Option(False)):
    try:
        success = upm.enable_disable_safemode(ctx.obj.get("base_url"), False, keep_state)
        if success:
            logging.info(
                f"Safe-mode is now {Fore.GREEN}disabled{Fore.RESET}, all plugins"
                f"{'got restored' if not keep_state else 'stayed disabled'}."
            )
        else:
            logging.error("Could not disable safe-mode - is safe-mode already disabled?")
    except requests.exceptions.ConnectionError:
        logging.error("Could not connect to host - check your base-url")
        sys.exit(1)
    except Exception as exc:
        logging.error("An error occured - check your credentials")
        logging.error("%s", exc)
        sys.exit(1)


@app_job.command("list")
def job_list(
    ctx: typer.Context,
    hide_default: typing.Optional[bool] = typer.Option(False),
    print_all_infos: typing.Optional[bool] = typer.Option(False),
):
    """ Confluence only, list all jobs available
    """
    try:
        logging.info("Getting jobs... This can take some time - please wait!")
        _job_list, token, cookies = jobs.list_jobs(ctx.obj.get("base_url"))

        columns, _ = shutil.get_terminal_size(fallback=(80, 24))

        width = int((columns - 17) / 4)
        print(
            f"{Fore.LIGHTBLACK_EX}{'idx':3} {'name':{width}} {'group':{width*2}} {'id':{width}} {'STS':3} {'RUNBL':5}{Fore.RESET}"
        )
        if print_all_infos:
            print(
                f"{Fore.LIGHTBLACK_EX}        {'last execution':{width}} {'next execution':{width}} {'avg duration':7}{Fore.RESET}"
            )
        print(f"{Fore.LIGHTBLACK_EX}{'='*columns}{Fore.RESET}")
        for idx, job in enumerate(_job_list):
            if hide_default and job.group == "DEFAULT":
                continue
            status_emoji = "🔄" if job.status == "Scheduled" else "❌"
            if job.is_runnable:
                runnable_emoji = f"{Fore.GREEN}✓{Fore.RESET}"
            else:
                runnable_emoji = f"{Fore.RED}!{Fore.RESET}"
            print(
                f"{Fore.YELLOW}{idx:3}{Fore.RESET} {job.name:{width}.{width}} {job.group:{width*2}.{width*2}}",
                f"{job.id:{width}.{width}} {status_emoji:2.6} {runnable_emoji}",
            )
            if print_all_infos:
                print(
                    f"        {job.last_execution:{width}.{width}} {job.next_execution:{width}.{width}}",
                    f"{job.avg_duration:{width}.{width}}",
                )
                print(f"{Fore.LIGHTBLACK_EX}{'-'*columns}{Fore.RESET}")
    except requests.exceptions.ConnectionError:
        logging.error("Could not connect to host - check your base-url")
        sys.exit(1)
    except Exception as exc:
        logging.error("An error occured - check your credentials")
        logging.error("%s", exc)
        sys.exit(1)


def _select_job(
    _job_list: typing.List[jobs.Job],
    idx: typing.Optional[int] = None,
    id: typing.Optional[str] = None,
    group: typing.Optional[str] = None,
) -> jobs.Job:
    selected_job = None

    if id is not None:
        possible_jobs = [x for x in _job_list if id in x.id]
        if len(possible_jobs) > 1 and group is not None:
            possible_jobs = [x for x in possible_jobs if group in x.group]

        if len(possible_jobs) == 1:
            selected_job = possible_jobs[0]
        elif len(possible_jobs) > 0:
            # TODO: ADD SELECTION HERE
            pass
    elif idx is not None:
        selected_job = _job_list[idx] if idx < len(_job_list) else None
    else:
        columns, _ = shutil.get_terminal_size(fallback=(80, 24))

        width = int((columns - 15) / 4)
        print(f"{Fore.LIGHTBLACK_EX}{'idx':3} {'name':{width}} {'group':{width*2}} {'id':{width}} runnable")
        for idx, job in enumerate(_job_list):
            if job.is_runnable:
                runnable_emoji = f"{Fore.GREEN}✓{Fore.RESET}"
            else:
                runnable_emoji = f"{Fore.RED}!{Fore.RESET}"
            print(
                f"{Fore.YELLOW}{idx:3}{Fore.RESET} {job.name:{width}.{width}} {job.group:{width*2}.{width*2}}",
                f"{job.id:{width}.{width}} {runnable_emoji}",
            )

        while True:
            _idx = typer.prompt("Select a job index (idx)")
            if int(_idx) >= 0 and int(_idx) < len(_job_list):
                idx = int(_idx)
                selected_job = _job_list[idx]
                break

    if selected_job is None:
        logging.error("job could not be found")
        typer.Exit(1)

    return selected_job


@app_job.command("run")
def job_run(
    ctx: typer.Context,
    idx: typing.Optional[int] = typer.Option(None),
    id: typing.Optional[str] = typer.Option(None),
    group: typing.Optional[str] = typer.Option(None),
):
    """ Confluence only, runs a specified job
    """
    try:
        logging.info("Getting jobs... This can take some time - please wait!")
        base_url = ctx.obj.get("base_url")
        _job_list, token, cookies = jobs.list_jobs(base_url)

        selected_job = _select_job(_job_list, idx, id, group)

        logging.info(f"Job {selected_job.name} ({selected_job.id}) selected - Trying to run the job now")

        success = jobs.run_job(base_url, selected_job, token, cookies)
        if success:
            logging.info("Started job successfully!")
        else:
            logging.error("Couldn't start job!")

    except requests.exceptions.ConnectionError:
        logging.error("Could not connect to host - check your base-url")
        sys.exit(1)
    except Exception as exc:
        logging.error("An error occured - check your credentials")
        logging.error("%s", exc)
        sys.exit(1)


@app_job.command("info")
def job_info(
    ctx: typer.Context,
    idx: typing.Optional[int] = typer.Option(None),
    id: typing.Optional[str] = typer.Option(None),
    group: typing.Optional[str] = typer.Option(None),
):
    try:
        logging.info("Getting jobs... This can take some time - please wait!")
        base_url = ctx.obj.get("base_url")
        _job_list, token, cookies = jobs.list_jobs(base_url)

        selected_job = _select_job(_job_list, idx, id, group)
        for key, value in selected_job.__dict__.items():
            print(f"{(key.replace('_', ' ') + ':'):25.25} {value}")
    except requests.exceptions.ConnectionError:
        logging.error("Could not connect to host - check your base-url")
        sys.exit(1)
    except Exception as exc:
        logging.error("An error occured - check your credentials")
        logging.error("%s", exc)
        sys.exit(1)


@app_job.command("disable")
def job_disable(
    ctx: typer.Context,
    idx: typing.Optional[int] = typer.Option(None),
    id: typing.Optional[str] = typer.Option(None),
    group: typing.Optional[str] = typer.Option(None),
):
    """ Confluence only, disable a specified job
    """
    try:
        logging.info("Getting jobs... This can take some time - please wait!")
        base_url = ctx.obj.get("base_url")
        _job_list, token, cookies = jobs.list_jobs(base_url)

        selected_job = _select_job(_job_list, idx, id, group)

        logging.info(f"Job {selected_job.name} ({selected_job.id}) selected - Trying to disable the job now")

        success = jobs.disable_job(base_url, selected_job, token, cookies)
        if success:
            logging.info("Disabled job successfully!")
        else:
            logging.error("Couldn't disable job!")
    except requests.exceptions.ConnectionError:
        logging.error("Could not connect to host - check your base-url")
        sys.exit(1)
    except Exception as exc:
        logging.error("An error occured - check your credentials")
        logging.error("%s", exc)
        sys.exit(1)


@app_job.command("enable")
def job_enable(
    ctx: typer.Context,
    idx: typing.Optional[int] = typer.Option(None),
    id: typing.Optional[str] = typer.Option(None),
    group: typing.Optional[str] = typer.Option(None),
):
    """ Confluence only, enable a specified job
    """
    try:
        logging.info("Getting jobs... This can take some time - please wait!")
        base_url = ctx.obj.get("base_url")
        _job_list, token, cookies = jobs.list_jobs(base_url)

        selected_job = _select_job(_job_list, idx, id, group)

        logging.info(f"Job {selected_job.name} ({selected_job.id}) selected - Trying to enable the job now")

        success = jobs.enable_job(base_url, selected_job, token, cookies)
        if success:
            logging.info("Enabled job successfully!")
        else:
            logging.error("Couldn't enable job!")
    except requests.exceptions.ConnectionError:
        logging.error("Could not connect to host - check your base-url")
        sys.exit(1)
    except Exception as exc:
        logging.error("An error occured - check your credentials")
        logging.error("%s", exc)
        sys.exit(1)


class TqdmUpTo(tqdm):
    """Provides `update_to(n)` which uses `tqdm.update(delta_n)`."""

    def update_to(self, b=1, bsize=1, tsize=None):
        """
        b  : int, optional
            Number of blocks transferred so far [default: 1].
        bsize  : int, optional
            Size of each block (in tqdm units) [default: 1].
        tsize  : int, optional
            Total size (in tqdm units). If [default: None] remains unchanged.
        """
        if tsize is not None:
            self.total = tsize
        self.update(b * bsize - self.n)  # will also set self.n = b * bsize


if __name__ == "__main__":
    main()
